from __future__ import annotations

import enum
import os
import textwrap
import warnings

import click

from .catalog import CUSTOM_SCHEMA_NAMES, SCHEMA_CATALOG
from .checker import SchemaChecker
from .formats import FormatOptions, RegexFormatBehavior
from .instance_loader import InstanceLoader
from .parsers import SUPPORTED_FILE_FORMATS
from .reporter import REPORTER_BY_NAME, Reporter
from .schema_loader import (
    BuiltinSchemaLoader,
    MetaSchemaLoader,
    SchemaLoader,
    SchemaLoaderBase,
)
from .transforms import TRANSFORM_LIBRARY, Transform

BUILTIN_SCHEMA_NAMES = [f"vendor.{k}" for k in SCHEMA_CATALOG.keys()] + [
    f"custom.{k}" for k in CUSTOM_SCHEMA_NAMES
]
BUILTIN_SCHEMA_CHOICES = (
    BUILTIN_SCHEMA_NAMES + list(SCHEMA_CATALOG.keys()) + CUSTOM_SCHEMA_NAMES
)

SHOWALL_DEPRECATION_MSG = """

  '--show-all-validation-errors' is deprecated and will be removed in a future version.

  Update your usage to use '-v/--verbose' instead.
"""


def evaluate_environment_settings(ctx: click.Context) -> None:
    if os.getenv("NO_COLOR") is not None:
        ctx.color = False


class SchemaLoadingMode(enum.Enum):
    filepath = "filepath"
    builtin = "builtin"
    metaschema = "metaschema"


class ParseResult:
    def __init__(self) -> None:
        # primary options: schema + instances
        self.schema_mode: SchemaLoadingMode = SchemaLoadingMode.filepath
        self.schema_path: str | None = None
        self.instancefiles: tuple[str, ...] = ()
        # cache controls
        self.disable_cache: bool = False
        self.cache_filename: str | None = None
        # filetype detection (JSON, YAML, TOML, etc)
        self.default_filetype: str = "json"
        # data-transform (for Azure Pipelines and potentially future transforms)
        self.data_transform: Transform | None = None
        # regex format options
        self.disable_format: bool = False
        self.format_regex: RegexFormatBehavior = RegexFormatBehavior.default
        # error and output controls
        self.verbosity: int = 1
        self.traceback_mode: str = "short"
        self.output_format: str = "text"

    def set_schema(
        self, schemafile: str | None, builtin_schema: str | None, check_metaschema: bool
    ) -> None:
        mutex_arg_count = sum(
            1 if x else 0 for x in (schemafile, builtin_schema, check_metaschema)
        )
        if mutex_arg_count == 0:
            raise click.UsageError(
                "Either --schemafile, --builtin-schema, or --check-metaschema "
                "must be provided"
            )
        if mutex_arg_count > 1:
            raise click.UsageError(
                "--schemafile, --builtin-schema, and --check-metaschema "
                "are mutually exclusive"
            )

        if schemafile:
            self.schema_mode = SchemaLoadingMode.filepath
            self.schema_path = schemafile
        elif builtin_schema:
            self.schema_mode = SchemaLoadingMode.builtin
            self.schema_path = builtin_schema
        else:
            self.schema_mode = SchemaLoadingMode.metaschema

    @property
    def format_opts(self) -> FormatOptions:
        return FormatOptions(
            enabled=not self.disable_format, regex_behavior=self.format_regex
        )


@click.command(
    "check-jsonschema",
    help="""\
Check JSON and YAML files against a JSON Schema.

The schema is specified either with '--schemafile' or with '--builtin-schema'.

'check-jsonschema' supports and checks the following formats by default:
    date, email, ipv4, regex, uuid

\b
For the "regex" format, there are multiple modes which can be specified with
'--format-regex':
    default  |  best effort check
    disabled |  do not check the regex format
    python   |  check that the string is a valid python regex

\b
The '--builtin-schema' flag supports the following schema names:
"""
    + textwrap.indent(
        "\n".join(
            textwrap.wrap(
                ", ".join(BUILTIN_SCHEMA_NAMES),
                width=75,
                break_long_words=False,
                break_on_hyphens=False,
            ),
        ),
        "    ",
    ),
)
@click.help_option("-h", "--help")
@click.version_option()
@click.option(
    "--schemafile",
    help=(
        "The path to a file containing the JSON Schema to use or an "
        "HTTP(S) URI for the schema. If a remote file is used, "
        "it will be downloaded and cached locally based on mtime."
    ),
)
@click.option(
    "--builtin-schema",
    help="The name of an internal schema to use for '--schemafile'",
    type=click.Choice(BUILTIN_SCHEMA_CHOICES, case_sensitive=False),
    metavar="BUILTIN_SCHEMA_NAME",
)
@click.option(
    "--check-metaschema",
    is_flag=True,
    help=(
        "Instead of validating the instances against a schema, treat each file as a "
        "schema and validate them under ther matching metaschemas."
    ),
)
@click.option(
    "--no-cache",
    is_flag=True,
    help="Disable schema caching. Always download remote schemas.",
)
@click.option(
    "--cache-filename",
    help=(
        "The name to use for caching a remote schema. "
        "Defaults to the last slash-delimited part of the URI."
    ),
)
@click.option(
    "--disable-format", is_flag=True, help="Disable all format checks in the schema."
)
@click.option(
    "--format-regex",
    help=(
        "Set the mode of format validation for regexes. "
        "If '--disable-format' is used, this option has no effect."
    ),
    default=RegexFormatBehavior.default.value,
    type=click.Choice([x.value for x in RegexFormatBehavior], case_sensitive=False),
)
@click.option(
    "--default-filetype",
    help="A default filetype to assume when a file's type is not detected",
    default="json",
    show_default=True,
    type=click.Choice(SUPPORTED_FILE_FORMATS, case_sensitive=True),
)
@click.option(
    "--traceback-mode",
    help=(
        "Set the mode of presentation for error traces. "
        "Defaults to shortened tracebacks."
    ),
    type=click.Choice(("full", "short")),
    default="short",
)
@click.option(
    "--data-transform",
    help=(
        "Select a builtin transform which should be applied to instancefiles before "
        "they are checked."
    ),
    type=click.Choice(tuple(TRANSFORM_LIBRARY.keys())),
)
@click.option(
    "-o",
    "--output-format",
    help="Which output format to use",
    type=click.Choice(tuple(REPORTER_BY_NAME.keys()), case_sensitive=False),
    default="text",
)
@click.option(
    "-v",
    "--verbose",
    help=(
        "Increase output verbosity. On validation errors, this may be especially "
        "useful when oneOf or anyOf is used in the schema."
    ),
    count=True,
)
@click.option(
    "-q",
    "--quiet",
    help="Reduce output verbosity",
    count=True,
)
@click.option(
    # TODO: remove in v0.15.1 or later
    "--show-all-validation-errors",
    is_flag=True,
    hidden=True,
)
@click.argument("instancefiles", required=True, nargs=-1)
@click.pass_context
def main(
    ctx: click.Context,
    *,
    schemafile: str | None,
    builtin_schema: str | None,
    check_metaschema: bool,
    no_cache: bool,
    cache_filename: str | None,
    disable_format: bool,
    format_regex: str,
    default_filetype: str,
    show_all_validation_errors: bool,
    traceback_mode: str,
    data_transform: str | None,
    output_format: str,
    verbose: int,
    quiet: int,
    instancefiles: tuple[str, ...],
) -> None:
    args = ParseResult()
    evaluate_environment_settings(ctx)

    args.set_schema(schemafile, builtin_schema, check_metaschema)
    args.instancefiles = instancefiles

    args.disable_format = disable_format
    args.format_regex = RegexFormatBehavior(format_regex)
    args.disable_cache = no_cache
    args.default_filetype = default_filetype
    if cache_filename is not None:
        args.cache_filename = cache_filename
    if data_transform is not None:
        args.data_transform = TRANSFORM_LIBRARY[data_transform]
    if show_all_validation_errors:
        warnings.warn(SHOWALL_DEPRECATION_MSG)

    # verbosity behavior:
    # - default is 1
    # - count '-v' with a min of 1 if --show-all-validation-errors was given
    # - subtract count of '-q'
    args.verbosity = 1 + max(verbose, 1 if show_all_validation_errors else 0) - quiet
    args.traceback_mode = traceback_mode
    args.output_format = output_format

    execute(args)


# separate parsing from execution for simpler mocking for unit tests


def build_schema_loader(args: ParseResult) -> SchemaLoaderBase:
    if args.schema_mode == SchemaLoadingMode.metaschema:
        return MetaSchemaLoader()
    elif args.schema_mode == SchemaLoadingMode.builtin:
        assert args.schema_path is not None
        return BuiltinSchemaLoader(args.schema_path)
    elif args.schema_mode == SchemaLoadingMode.filepath:
        assert args.schema_path is not None
        return SchemaLoader(args.schema_path, args.cache_filename, args.disable_cache)
    else:
        raise NotImplementedError("no valid schema option provided")


def build_instance_loader(args: ParseResult) -> InstanceLoader:
    return InstanceLoader(
        args.instancefiles,
        default_filetype=args.default_filetype,
        data_transform=args.data_transform,
    )


def build_reporter(args: ParseResult) -> Reporter:
    cls = REPORTER_BY_NAME[args.output_format]
    return cls(verbosity=args.verbosity)


def build_checker(args: ParseResult) -> SchemaChecker:
    schema_loader = build_schema_loader(args)
    instance_loader = build_instance_loader(args)
    reporter = build_reporter(args)
    return SchemaChecker(
        schema_loader,
        instance_loader,
        reporter,
        format_opts=args.format_opts,
        traceback_mode=args.traceback_mode,
    )


def execute(args: ParseResult) -> None:
    checker = build_checker(args)
    ret = checker.run()
    click.get_current_context().exit(ret)



from __future__ import annotations

import pathlib
import typing as t

from .parsers import ParseError, ParserSet
from .transforms import Transform


class InstanceLoader:
    def __init__(
        self,
        filenames: t.Sequence[str],
        default_filetype: str = "json",
        data_transform: Transform | None = None,
    ) -> None:
        self._filenames = filenames
        self._default_filetype = default_filetype
        self._data_transform = (
            data_transform if data_transform is not None else Transform()
        )

        self._parsers = ParserSet(
            modify_yaml_implementation=self._data_transform.modify_yaml_implementation
        )

    def iter_files(self) -> t.Iterator[tuple[pathlib.Path, ParseError | t.Any]]:
        for fn in self._filenames:
            path = pathlib.Path(fn)
            try:
                data: t.Any = self._parsers.parse_file(path, self._default_filetype)
            except ParseError as err:
                data = err
            else:
                data = self._data_transform(data)
            yield (path, data)



from __future__ import annotations

import datetime
import typing as t

try:
    import tomli

    has_toml = True
except ImportError:
    has_toml = False


def _normalize(data: t.Any) -> t.Any:
    """
    Normalize TOML data to fit the requirements to be JSON-encodeable.

    Currently this applies the following transformations:

        offset-aware datetime.datetime values are converted to strings using isoformat()
        naive datetime.datetime values are converted to strings using isoformat() + "Z"

        offset-aware datetime.time values are converted to strings using isoformat()
        naive datetime.time values are converted to strings using isoformat() + "Z"

        datetime.date values are converted to strings using isoformat()
    """
    if isinstance(data, dict):
        return {k: _normalize(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_normalize(x) for x in data]
    else:
        # python's datetime will format to an ISO partial time when handling a naive
        # time/datetime , but JSON Schema format validation specifies that date-time is
        # taken from RFC3339, which defines "date-time" as including 'Z|offset'
        # the specification for "time" is less clear because JSON Schema does not specify
        # which RFC3339 definition should be used, and the RFC has no format named "time",
        # only "full-time" (with Z|offset) and "partial-time" (no offset)
        #
        # rfc3339_validator (used by 'jsonschema') requires the offset, so we will do the
        # same
        if isinstance(data, datetime.datetime) or isinstance(data, datetime.time):
            if data.tzinfo is None:
                return data.isoformat() + "Z"
            return data.isoformat()
        elif isinstance(data, datetime.date):
            return data.isoformat()
        return data


# present a bool for detecting that it's enabled
ENABLED = has_toml


if has_toml:
    ParseError: type[Exception] = tomli.TOMLDecodeError

    def load(stream: t.BinaryIO) -> t.Any:
        data = tomli.load(stream)
        return _normalize(data)

else:
    ParseError = ValueError

    def load(stream: t.BinaryIO) -> t.Any:
        raise NotImplementedError


MISSING_SUPPORT_MESSAGE = """
check-jsonschema can only parse TOML files when a TOML parser is installed

If you are running check-jsonschema as an installed python package, add support with
    pip install tomli

If you are running check-jsonschema as a pre-commit hook, set
    additional_dependencies: ['tomli']
"""



from __future__ import annotations

import typing as t

# try to import pyjson5 first
# this is the CPython implementation and therefore preferred for its speec
try:
    import pyjson5

    ParseError: type[Exception] = pyjson5.Json5DecoderException
    _load: t.Callable | None = pyjson5.load
except ImportError:
    # if pyjson5 was not available, try to import 'json5', the pure-python implementation
    try:
        import json5

        # json5 doesn't define a custom decoding error class
        ParseError = ValueError
        _load = json5.load
    except ImportError:
        ParseError = ValueError
        _load = None

# present a bool for detecting that it's enabled
ENABLED = _load is not None

if _load is not None:
    _load_concrete: t.Callable = _load

    def load(stream: t.BinaryIO) -> t.Any:
        return _load_concrete(stream)

else:

    def load(stream: t.BinaryIO) -> t.Any:
        raise NotImplementedError


MISSING_SUPPORT_MESSAGE = """
check-jsonschema can only parse json5 files when a json5 parser is installed

If you are running check-jsonschema as an installed python package, either
    pip install json5
or
    pip install pyjson5

If you are running check-jsonschema as a pre-commit hook, set
    additional_dependencies: ['json5']
or
    additional_dependencies: ['pyjson5']
"""



from __future__ import annotations

import typing as t


def _githubusercontent_url(owner: str, repo: str, ref: str, path: str) -> str:
    return f"https://raw.githubusercontent.com/{owner}/{repo}/{ref}/{path}"


# this lists custom schemas which are *not* part of the catalog
CUSTOM_SCHEMA_NAMES = [
    "github-workflows-require-timeout",
]

# Known configs. The SchemaCatalog lists known schema URLs with their names.
# kept in alphabetical order by name
#
# Additional config could be associated with the schemas in the future.
SCHEMA_CATALOG: dict[str, dict[str, t.Any]] = {
    "azure-pipelines": {
        "url": _githubusercontent_url(
            "microsoft", "azure-pipelines-vscode", "main", "service-schema.json"
        ),
        "hook_config": {
            "name": "Validate Azure Pipelines",
            "description": "Validate Azure Pipelines config against the schema provided "
            "by Microsoft",
            "add_args": ["--data-transform", "azure-pipelines"],
            "files": r"^(\.)?azure-pipelines.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "bamboo-spec": {
        "url": "https://json.schemastore.org/bamboo-spec.json",
        "hook_config": {
            "name": "Validate Bamboo Specs",
            "files": r"^bamboo-specs/.*\.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "dependabot": {
        "url": "https://json.schemastore.org/dependabot-2.0.json",
        "hook_config": {
            "name": "Validate Dependabot Config (v2)",
            "files": r"^\.github/dependabot.yml$",
            "types": "yaml",
        },
    },
    "github-actions": {
        "url": "https://json.schemastore.org/github-action",
        "hook_config": {
            "name": "Validate GitHub Actions",
            "files": ["action.(yml|yaml)", r"\.github/actions/.*"],
            "types": "yaml",
        },
    },
    "github-workflows": {
        "url": "https://json.schemastore.org/github-workflow",
        "hook_config": {
            "name": "Validate GitHub Workflows",
            "files": r"^\.github/workflows/[^/]+$",
            "types": "yaml",
        },
    },
    "gitlab-ci": {
        "url": "https://gitlab.com/gitlab-org/gitlab/-/raw/master/app/assets/javascripts"
        "/editor/schema/ci.json",
        "hook_config": {
            "name": "Validate GitLab CI config",
            "add_args": ["--data-transform", "gitlab-ci"],
            "files": r"^.*\.gitlab-ci.yml$",
            "types": "yaml",
        },
    },
    "readthedocs": {
        "url": _githubusercontent_url(
            "readthedocs",
            "readthedocs.org",
            "master",
            "readthedocs/rtd_tests/fixtures/spec/v2/schema.json",
        ),
        "hook_config": {
            "name": "Validate ReadTheDocs Config",
            "description": "Validate ReadTheDocs config against the schema "
            "provided by ReadTheDocs",
            "files": r"^\.readthedocs.(yml|yaml)$",
            "types": "yaml",
        },
    },
    "renovate": {
        "url": "https://docs.renovatebot.com/renovate-schema.json",
        "hook_config": {
            "name": "Validate Renovate Config",
            "description": "Validate Renovate config against the schema provided by "
            "Renovate (does not support renovate config in package.json)",
            "files": [
                r"renovate\.(json|json5)",
                r"\.(github|gitlab)/renovate\.(json|json5)",
                r"\.renovaterc(\.json)?",
            ],
        },
    },
    "travis": {
        "url": "https://json.schemastore.org/travis",
        "hook_config": {
            "name": "Validate Travis Config",
            "files": r"^\.travis.(yml|yaml)$",
            "types": "yaml",
        },
    },
}



from __future__ import annotations

try:
    # first, try to import the installed package version
    import importlib_resources
except ImportError:
    # if it's not installed, assume that the stdlib version is new enough (e.g. py3.10)
    import importlib.resources as importlib_resources  # type: ignore[no-redef]

import json
import typing as t


class NoSuchSchemaError(ValueError):
    pass


def _get(package: str, resource: str, name: str) -> dict[str, t.Any]:
    try:
        return t.cast(
            "dict[str, t.Any]",
            json.loads(
                importlib_resources.files(package).joinpath(resource).read_bytes()
            ),
        )
    except (FileNotFoundError, ModuleNotFoundError):
        raise NoSuchSchemaError(f"no builtin schema named {name} was found")


def _get_vendored_schema(name: str) -> dict[str, t.Any]:
    return _get("check_jsonschema.builtin_schemas.vendor", f"{name}.json", name)


def _get_custom_schema(name: str) -> dict[str, t.Any]:
    return _get("check_jsonschema.builtin_schemas.custom", f"{name}.json", name)


def get_builtin_schema(name: str) -> dict[str, t.Any]:
    # first, look for an identifying prefix
    if name.startswith("vendor."):
        return _get_vendored_schema(name[7:])
    elif name.startswith("custom."):
        return _get_custom_schema(name[7:])

    # if there is no prefix, just try in order: first custom, then vendored
    try:
        return _get_custom_schema(name)
    except NoSuchSchemaError:
        return _get_vendored_schema(name)



from __future__ import annotations

import typing as t

import click
import jsonschema


class _CliRefResolver(jsonschema.RefResolver):
    def resolve_remote(self, uri: str) -> t.Any:
        if uri.endswith(".yaml") or uri.endswith(".yml"):
            click.secho(
                """\
WARNING: You appear to be using a schema which references a YAML file.

This is not supported by check-jsonschema and may result in errors.
""",
                err=True,
                fg="yellow",
            )
        elif uri.endswith(".json5"):
            click.secho(
                """\
WARNING: You appear to be using a schema which references a JSON5 file.

This is not supported by check-jsonschema and may result in errors.
""",
                err=True,
                fg="yellow",
            )
        return super().resolve_remote(uri)


def make_ref_resolver(
    schema_uri: str | None, schema: dict
) -> jsonschema.RefResolver | None:
    if not schema_uri:
        return None

    base_uri = schema.get("$id", schema_uri)
    # FIXME: temporary type-ignore beause typeshed has the type wrong
    return _CliRefResolver(base_uri, schema)  # type: ignore[arg-type]



from __future__ import annotations

import pathlib
import typing as t
import urllib.error
import urllib.parse

import jsonschema

from ..builtin_schemas import get_builtin_schema
from ..formats import FormatOptions, make_format_checker
from ..utils import is_url_ish
from .errors import UnsupportedUrlScheme
from .readers import HttpSchemaReader, LocalSchemaReader
from .resolver import make_ref_resolver


class SchemaLoaderBase:
    def get_validator(
        self,
        path: pathlib.Path,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
    ) -> jsonschema.Validator:
        raise NotImplementedError


class SchemaLoader(SchemaLoaderBase):
    def __init__(
        self,
        schemafile: str,
        cache_filename: str | None = None,
        disable_cache: bool = False,
    ) -> None:
        # record input parameters (these are not to be modified)
        self.schemafile = schemafile
        self.cache_filename = cache_filename
        self.disable_cache = disable_cache

        # if the schema location is a URL, which may include a file:// URL, parse it
        self.url_info = None
        if is_url_ish(self.schemafile):
            self.url_info = urllib.parse.urlparse(self.schemafile)

        # setup a schema reader lazily, when needed
        self._reader: LocalSchemaReader | HttpSchemaReader | None = None

        # setup a location to store the validator so that it is only built once by default
        self._validator: jsonschema.Validator | None = None

    @property
    def reader(self) -> LocalSchemaReader | HttpSchemaReader:
        if self._reader is None:
            self._reader = self._get_schema_reader()
        return self._reader

    def _get_schema_reader(self) -> LocalSchemaReader | HttpSchemaReader:
        if self.url_info is None or self.url_info.scheme in ("file", ""):
            return LocalSchemaReader(self.schemafile)

        if self.url_info.scheme in ("http", "https"):
            return HttpSchemaReader(
                self.schemafile,
                self.cache_filename,
                self.disable_cache,
            )
        else:
            raise UnsupportedUrlScheme(
                "check-jsonschema only supports http, https, and local files. "
                f"detected parsed URL had an unrecognized scheme: {self.url_info}"
            )

    def get_schema_ref_base(self) -> str | None:
        return self.reader.get_ref_base()

    def get_schema(self) -> dict[str, t.Any]:
        return self.reader.read_schema()

    def make_validator(self, format_opts: FormatOptions) -> jsonschema.Validator:
        schema_uri = self.get_schema_ref_base()
        schema = self.get_schema()

        schema_dialect = schema.get("$schema")

        # format checker (which may be None)
        format_checker = make_format_checker(format_opts, schema_dialect)

        # ref resolver which may be built from the schema path
        # if the location is a URL, there's no change, but if it's a file path
        # it's made absolute and URI-ized
        # the resolver should use `$id` if there is one present in the schema
        ref_resolver = make_ref_resolver(schema_uri, schema)

        # get the correct validator class and check the schema under its metaschema
        validator_cls = jsonschema.validators.validator_for(schema)
        validator_cls.check_schema(schema)

        # now that we know it's safe to try to create the validator instance, do it
        validator = validator_cls(
            schema,
            resolver=ref_resolver,
            format_checker=format_checker,
        )
        return t.cast(jsonschema.Validator, validator)

    def get_validator(
        self,
        path: pathlib.Path,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
    ) -> jsonschema.Validator:
        self._validator = self.make_validator(format_opts)
        return self._validator


class BuiltinSchemaLoader(SchemaLoader):
    def __init__(self, schema_name: str) -> None:
        self.schema_name = schema_name

    def get_schema_ref_base(self) -> str | None:
        return None

    def get_schema(self) -> dict[str, t.Any]:
        return get_builtin_schema(self.schema_name)


class MetaSchemaLoader(SchemaLoaderBase):
    def get_validator(
        self,
        path: pathlib.Path,
        instance_doc: dict[str, t.Any],
        format_opts: FormatOptions,
    ) -> jsonschema.Validator:
        validator = jsonschema.validators.validator_for(instance_doc)
        return t.cast(jsonschema.Validator, validator(validator.META_SCHEMA))



"""
A data transform which unpacks "compile-time expressions" from Azure Pipelines files.

For the original source which inspired this transform, see the YAML parser used in the
Azure Pipelines Language Server:
    https://github.com/microsoft/azure-pipelines-language-server/blob/71b20f92874c02dfe82ad2cc2dcc7fa64996be91/language-service/src/parser/yamlParser.ts#L182

That source is licensed under the MIT License.
The original license can be found in
    src/check_jsonschema/builtin_schemas/vendor/LICENSE.azure-pipelines


The transform does not deeply interpret the expressions. It just "unnests" them.

It will turn this input

    jobs:
      - ${{ each val in parameter.vals }}:
        - job: foo
          steps:
            - bash: echo ${{ val }}

into

    jobs:
      - job: foo
        steps:
          - bash: echo ${{ val }}
"""
from __future__ import annotations

import typing as t

from .base import Transform


class AzurePipelinesDataError(ValueError):
    def __init__(self, message: str) -> None:
        super().__init__(f"azure-pipelines transform: {message}")


def is_expression(s: str) -> bool:
    return s.startswith("${{") and s.endswith("}}")


def traverse_data(data: t.Any) -> t.Any:
    if isinstance(data, dict):
        return traverse_dict(data)
    if isinstance(data, list):
        return traverse_list(data)
    return data


def traverse_list(data: list) -> list:
    ret = []
    for item in data:
        # is the current item a single-value dict with an expression as its key?
        item_is_expr = (
            isinstance(item, dict)
            and len(item) == 1
            and is_expression(tuple(item)[0])  # tuple() gets keys
        )

        if item_is_expr:
            # unpack the expression item and recurse over the value
            item_key, item_value = list(item.items())[0]
            item_value = traverse_data(item_value)

            if isinstance(item_value, list):
                ret.extend(item_value)
            elif isinstance(item_value, dict):
                ret.append(item_value)
            else:
                ret.append(item_value)
        # not expression? process the item and append
        else:
            ret.append(traverse_data(item))
    return ret


def traverse_dict(data: dict) -> dict:
    newdata = {}
    for key, value in data.items():
        newvalue = traverse_data(value)
        if is_expression(key):
            # WARNING -- correctness unclear
            #
            # "lift" any dict by moving its attributes up into the object being evaluated
            #
            # e.g.
            #    parent:
            #      ${{ each x in xs }}:
            #        - k: v-${{ x }}
            #
            # becomes
            #
            #    parent:
            #      - k: v-${{ x }}
            if isinstance(newvalue, dict):
                for add_k, add_v in newvalue.items():
                    newdata[add_k] = add_v
            # In all other cases, drop the content from the data. This is based on the
            # azure-pipelines-language server behavior:
            # https://github.com/microsoft/azure-pipelines-language-server/blob/71b20f92874c02dfe82ad2cc2dcc7fa64996be91/language-service/src/parser/yamlParser.ts#L185
            #
            # earlier versions would raise an error here, but this caused issues with
            # data in which expressions were mapped to simple strings
            #
            # e.g.
            #
            #    parent:
            #      ${{ x }}: ${{ y }}
            #
            # which occurs naturally *after* a lifting operation, as in
            #
            #    parent:
            #      ${{ each x, y in attrs }}:
            #        ${{ x }}: ${{ y }}
            else:
                continue
        else:
            newdata[key] = newvalue
    return newdata


def azure_main(data: dict | list) -> dict | list:
    if isinstance(data, list):
        raise AzurePipelinesDataError(
            "this transform requires that the data be an object, got list"
        )
    return traverse_dict(data)


AZURE_TRANSFORM = Transform(on_data=azure_main)



from __future__ import annotations

import json
import typing as t

import ruamel.yaml

from ..cachedownloader import CacheDownloader
from ..parsers import ParserSet
from ..utils import filename2path
from .errors import SchemaParseError

yaml = ruamel.yaml.YAML(typ="safe")


def _run_load_callback(schema_location: str, callback: t.Callable) -> dict:
    try:
        schema = callback()
    # only local loads can raise the YAMLError, but catch for both cases for simplicity
    except (ValueError, ruamel.yaml.error.YAMLError) as e:
        raise SchemaParseError(schema_location) from e
    if not isinstance(schema, dict):
        raise SchemaParseError(schema_location)
    return schema


class LocalSchemaReader:
    FORMATS = ("json", "json5", "yaml")

    def __init__(self, filename: str) -> None:
        self.path = filename2path(filename)
        self.filename = str(self.path)
        self.parsers = ParserSet(supported_formats=self.FORMATS)

    def get_ref_base(self) -> str:
        return self.path.as_uri()

    def _read_impl(self) -> t.Any:
        return self.parsers.parse_file(self.path, default_filetype="json")

    def read_schema(self) -> dict:
        return _run_load_callback(self.filename, self._read_impl)


class HttpSchemaReader:
    def __init__(
        self,
        url: str,
        cache_filename: str | None,
        disable_cache: bool,
    ) -> None:
        self.url = url
        self.downloader = CacheDownloader(
            url, cache_filename, disable_cache=disable_cache
        )

    def get_ref_base(self) -> str:
        return self.url

    def _read_impl(self) -> t.Any:
        with self.downloader.open() as fp:
            return json.load(fp)

    def read_schema(self) -> dict:
        return _run_load_callback(self.url, self._read_impl)



from __future__ import annotations

import copy
import enum
import re
import typing as t

import jsonschema

CHECKERS_BY_DIALECT = {
    "http://json-schema.org/draft-03/schema#": jsonschema.Draft3Validator.FORMAT_CHECKER,
    "http://json-schema.org/draft-04/schema#": jsonschema.Draft4Validator.FORMAT_CHECKER,
    "http://json-schema.org/draft-06/schema#": jsonschema.Draft6Validator.FORMAT_CHECKER,
    "http://json-schema.org/draft-07/schema#": jsonschema.Draft7Validator.FORMAT_CHECKER,
    "http://json-schema.org/draft-2019-09/schema": (
        jsonschema.Draft201909Validator.FORMAT_CHECKER
    ),
    "http://json-schema.org/draft-2020-12/schema": (
        jsonschema.Draft202012Validator.FORMAT_CHECKER
    ),
}
LATEST_DIALECT = "http://json-schema.org/draft-2020-12/schema"


def _regex_check(instance: t.Any) -> bool:
    if not isinstance(instance, str):
        return True
    re.compile(instance)
    return True


def _gated_regex_check(instance: t.Any) -> bool:
    if not isinstance(instance, str):
        return True
    if re.search(r"\(\?[^!=]", instance):
        return True
    re.compile(instance)
    return True


class RegexFormatBehavior(enum.Enum):
    default = "default"
    disabled = "disabled"
    python = "python"


class FormatOptions:
    def __init__(
        self,
        *,
        enabled: bool = True,
        regex_behavior: RegexFormatBehavior = RegexFormatBehavior.default,
    ) -> None:
        self.enabled = enabled
        self.regex_behavior = regex_behavior


def get_base_format_checker(schema_dialect: str | None) -> jsonschema.FormatChecker:
    # map a schema's `$schema` attribute (the dialect / JSON Schema version) to a matching
    # format checker
    # default to the latest draft
    if schema_dialect is None or schema_dialect not in CHECKERS_BY_DIALECT:
        return CHECKERS_BY_DIALECT[LATEST_DIALECT]
    return CHECKERS_BY_DIALECT[schema_dialect]


def make_format_checker(
    opts: FormatOptions,
    schema_dialect: str | None = None,
) -> jsonschema.FormatChecker | None:
    if not opts.enabled:
        return None

    # copy the base checker
    base_checker = get_base_format_checker(schema_dialect)
    checker = copy.deepcopy(base_checker)

    # remove the regex check
    del checker.checkers["regex"]

    if opts.regex_behavior == RegexFormatBehavior.disabled:
        pass
    elif opts.regex_behavior == RegexFormatBehavior.default:
        checker.checks("regex", raises=re.error)(_gated_regex_check)
    elif opts.regex_behavior == RegexFormatBehavior.python:
        checker.checks("regex", raises=re.error)(_regex_check)
    else:  # pragma: no cover
        raise NotImplementedError

    return checker
